/*  Basic m68hc05 simulator
    Copyright 2004 Free Software Foundation, Inc.
    Copyright 2004 Martin Kgler
    Contributed by Martin Kgler (mkoegler@auto.tuwien.ac.at)
    (based on m68hc11-run)
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
    
    */

#include "sim-main.h"
#include "sim-hw.h"
#include "hw-main.h"
#include "sim-options.h"
#include "hw-base.h"
#include <limits.h>

/* DEVICE

        m68hc05cpu - m68hc05 cpu virtual device

   PROPERTIES

   clock <hz>

        Frequency of the quartz used by the processor.

  PORTS

   reset (input)

        Reset the cpu and generates a cpu-reset event (used to reset
        other devices).

   cpu-reset (output)

        Event generated after the CPU performs a reset.


*/

struct m68hc05cpu
{
  /* Pending interrupts for delivery by event handler.  */
  int pending_reset;
  struct hw_event *event;
};


static void
dv_m68hc05_attach_address_callback (struct hw *me,
				    int level,
				    int space,
				    address_word addr,
				    address_word nr_bytes, struct hw *client)
{
  HW_TRACE ((me, "attach - level=%d, space=%d, addr=0x%lx, sz=%ld, client=%s",
	     level, space, (unsigned long) addr, (unsigned long) nr_bytes,
	     hw_path (client)));

  if (space != io_map)
    {
      sim_core_attach (hw_system (me), NULL,	/*cpu */
		       level, access_read_write_exec, space, addr, nr_bytes, 0,	/* modulo */
		       client, NULL);
    }
  else
    {
      /*printf("Attach from sub device: %d\n", (long) addr); */
      sim_core_attach (hw_system (me), NULL,	/*cpu */
		       level, access_io, space, addr, nr_bytes, 0,	/* modulo */
		       client, NULL);
    }
}

static void
dv_m68hc05_detach_address_callback (struct hw *me,
				    int level,
				    int space,
				    address_word addr,
				    address_word nr_bytes, struct hw *client)
{
  sim_core_detach (hw_system (me), NULL,	/*cpu */
		   level, space, addr);
}

/* An event arrives on an interrupt port.  */

static void
deliver_m68hc05cpu_interrupt (struct hw *me, void *data)
{
}

static void
m68hc05_delete (struct hw *me)
{
  struct m68hc05cpu *controller;

  controller = hw_data (me);

}



static const OPTION m68hc05_options[] = {
  {{NULL, no_argument, NULL, 0}, '\0', NULL, NULL, NULL}
};


/* input port ID's */

enum
{
  RESET_PORT,
  CPU_RESET_PORT,
  DUMMY_PORT,
};


static const struct hw_port_descriptor m68hc05cpu_ports[] = {

  /* Interrupt inputs.  */
  {"reset", RESET_PORT, 0, input_port,},

  /* Events generated for connection to other devices.  */
  {"cpu-reset", CPU_RESET_PORT, 0, output_port,},

  /* Interrupt inputs.  */
  {"dummy", DUMMY_PORT, 0, input_port,},

  {NULL,},
};



static void
m68hc05cpu_port_event (struct hw *me,
		       int my_port,
		       struct hw *source, int source_port, int level)
{
  struct m68hc05cpu *controller = hw_data (me);
  SIM_DESC sd;
  sim_cpu *cpu;

  sd = hw_system (me);
  cpu = STATE_CPU (sd, 0);
  switch (my_port)
    {
    case RESET_PORT:
      HW_TRACE ((me, "port-in reset"));

      cpu_reset (cpu);
      hw_port_event (me, CPU_RESET_PORT, 1);
      cpu_restart (cpu);
      break;
    case DUMMY_PORT:
      break;
    default:
      hw_abort (me, "bad switch");
      break;
    }

  /* Schedule an event to be delivered immediately after current
     instruction.  */
  if (controller->event != NULL)
    hw_event_queue_deschedule (me, controller->event);
  controller->event =
    hw_event_queue_schedule (me, 0, deliver_m68hc05cpu_interrupt, NULL);
}

static void
attach_m68hc05_regs (struct hw *me, struct m68hc05cpu *controller)
{
  SIM_DESC sd;
  sim_cpu *cpu;
  reg_property_spec reg;
  const char *cpu_mode;

  set_hw_delete (me, m68hc05_delete);

  /* Get cpu frequency.  */
  sd = hw_system (me);
  cpu = STATE_CPU (sd, 0);
  if (hw_find_property (me, "clock") != NULL)
    {
      cpu->cpu_frequency = hw_find_integer_property (me, "clock");
    }
  else
    {
      cpu->cpu_frequency = 8 * 1000 * 1000;
    }

  /* Add oscillator commands.  */
  sim_add_option_table (sd, 0, m68hc05_options);
}



static void
m68hc05cpu_finish (struct hw *me)
{
  struct m68hc05cpu *controller;

  controller = HW_ZALLOC (me, struct m68hc05cpu);
  set_hw_data (me, controller);
  set_hw_ports (me, m68hc05cpu_ports);
  set_hw_port_event (me, m68hc05cpu_port_event);
  set_hw_attach_address (me, dv_m68hc05_attach_address_callback);
  set_hw_detach_address (me, dv_m68hc05_detach_address_callback);

  /* Initialize the pending interrupt flags.  */
  controller->pending_reset = 0;
  controller->event = NULL;

  attach_m68hc05_regs (me, controller);
}

const struct hw_descriptor dv_m68hc05_descriptor[] = {
  {"m68hc05", m68hc05cpu_finish},
  {NULL},
};
