/*  Basic m68hc05 simulator
    Copyright 2004 Free Software Foundation, Inc.
    Copyright 2004 Martin Kgler
    Contributed by Martin Kgler (mkoegler@auto.tuwien.ac.at)
    (based on m68hc11-run)
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
    
    */

#ifndef _SIM_MAIN_H
#define _SIM_MAIN_H

#include "sim-basics.h"

typedef address_word sim_cia;

#include "sim-base.h"

#include "gdb/callback.h"
#include "gdb/remote-sim.h"
#include "sim-types.h"

#include "bfd.h"


typedef unsigned8 uint8;
typedef unsigned16 uint16;
typedef signed16 int16;
typedef unsigned32 uint32;
typedef signed32 int32;
typedef unsigned64 uint64;
typedef signed64 int64;

#include "interrupts.h"

#define A_REGNUM        0
#define X_REGNUM        1
#define SP_REGNUM       2
#define PC_REGNUM       3
#define PSW_REGNUM      4
#define CARRY_REGNUM    5



#define M68HC05_H_BIT 0x10
#define M68HC05_I_BIT 0x08
#define M68HC05_N_BIT 0x04
#define M68HC05_Z_BIT 0x02
#define M68HC05_C_BIT 0x01


typedef struct m68hc05_regs
{
  unsigned char a;
  unsigned char x;
  unsigned char ccr;
  unsigned char sp;
  unsigned short pc;
} m68hc05_regs;


typedef void (*cpu_interp) (struct _sim_cpu *);

struct _sim_cpu
{
  /* CPU registers.  */
  struct m68hc05_regs cpu_regs;
  /* CPU interrupts.  */
  struct interrupts cpu_interrupts;

  /* Pointer to the interpretor routine.  */
  cpu_interp cpu_interpretor;

  /* CPU absolute cycle time.  The cycle time is updated after
     each instruction, by the number of cycles taken by the instruction.
     It is cleared only when reset occurs.  */
  signed64 cpu_absolute_cycle;

  /* Number of cycles to increment after the current instruction.
     This is also the number of ticks for the generic event scheduler.  */
  uint8 cpu_current_cycle;
  int cpu_emul_syscall;
  int cpu_is_initialized;
  int cpu_running;
  int cpu_check_memory;
  int cpu_stop_on_interrupt;

  uint16 cpu_insn_pc;

  /* CPU frequency.  This is the quartz frequency.  It is divided by 4 to
     get the cycle time.  This is used for the timer rate and for the baud
     rate generation.  */
  unsigned long cpu_frequency;

  /* When this is set, start execution of program at address specified
     in the ELF header.  This is used for testing some programs that do not
     have an interrupt table linked with them.  Programs created during the
     GCC validation are like this. A normal 68HC11 does not behave like
     this (unless there is some OS or downloadable feature).  */
  int cpu_use_elf_start;

  /* The starting address specified in ELF header.  */
  int cpu_elf_start;

  struct hw *hw_cpu;

  /* ... base type ... */
  sim_cpu_base base;

};


#define cpu_get_a(PROC)          ((PROC)->cpu_regs.a)
#define cpu_get_x(PROC)          ((PROC)->cpu_regs.x)
#define cpu_get_ccr(PROC)        ((PROC)->cpu_regs.ccr)
#define cpu_get_sp(PROC)         ((PROC)->cpu_regs.sp)
#define cpu_get_pc(PROC)         ((PROC)->cpu_regs.pc)

#define cpu_set_a(PROC,VAL)      ((PROC)->cpu_regs.a=(VAL))
#define cpu_set_x(PROC,VAL)      ((PROC)->cpu_regs.x=(VAL))
#define cpu_set_ccr(PROC,VAL)    ((PROC)->cpu_regs.ccr=(VAL))
#define cpu_set_sp(PROC,VAL)     ((PROC)->cpu_regs.sp=(VAL))
#define cpu_set_pc(PROC,VAL)     ((PROC)->cpu_regs.pc=(VAL))

#define cpu_get_ccr_H(PROC)      ((cpu_get_ccr(PROC) & M68HC05_H_BIT) ? 1 : 0)
#define cpu_get_ccr_I(PROC)      ((cpu_get_ccr(PROC) & M68HC05_I_BIT) ? 1 : 0)
#define cpu_get_ccr_N(PROC)      ((cpu_get_ccr(PROC) & M68HC05_N_BIT) ? 1 : 0)
#define cpu_get_ccr_Z(PROC)      ((cpu_get_ccr(PROC) & M68HC05_Z_BIT) ? 1 : 0)
#define cpu_get_ccr_C(PROC)      ((cpu_get_ccr(PROC) & M68HC05_C_BIT) ? 1 : 0)

#define cpu_set_ccr_flag(S,B,V) \
cpu_set_ccr(S,(cpu_get_ccr(S) & ~(B)) | ((V) ? B : 0))

#define cpu_set_ccr_H(PROC,VAL)    cpu_set_ccr_flag(PROC, M68HC05_H_BIT, VAL)
#define cpu_set_ccr_I(PROC,VAL)    cpu_set_ccr_flag(PROC, M68HC05_I_BIT, VAL)
#define cpu_set_ccr_N(PROC,VAL)    cpu_set_ccr_flag(PROC, M68HC05_N_BIT, VAL)
#define cpu_set_ccr_Z(PROC,VAL)    cpu_set_ccr_flag(PROC, M68HC05_Z_BIT, VAL)
#define cpu_set_ccr_C(PROC,VAL)    cpu_set_ccr_flag(PROC, M68HC05_C_BIT, VAL)

#define cpu_add_cycles(PROC,T)  ((PROC)->cpu_current_cycle += (signed64) (T))

/* The current state of the processor; registers, memory, etc.  */

#define CIA_GET(CPU)      (cpu_get_pc (CPU))
#define CIA_SET(CPU,VAL)  (cpu_set_pc ((CPU), (VAL)))

#if (WITH_SMP)
#define STATE_CPU(sd,n) (&(sd)->cpu[n])
#else
#define STATE_CPU(sd,n) (&(sd)->cpu[0])
#endif

struct sim_state
{
  sim_cpu cpu[MAX_NR_PROCESSORS];
  device *devices;
  sim_state_base base;
};

extern void cpu_info (SIM_DESC sd, sim_cpu * proc);
extern int cpu_initialize (SIM_DESC sd, sim_cpu * cpu);
extern void cpu_single_step (sim_cpu * proc);
extern int cpu_reset (sim_cpu * cpu);
extern int cpu_restart (sim_cpu * cpu);
extern void cpu_interp_m68hc05 (sim_cpu * cpu);
extern void cpu_special (sim_cpu * proc, uint8 special);
extern void sim_board_reset (SIM_DESC sd);

#define PRINT_TIME  0x01
#define PRINT_CYCLE 0x02
extern const char *cycle_to_string (sim_cpu * cpu, signed64 t, int flags);
extern void cpu_memory_exception (struct _sim_cpu *proc,
				  SIM_SIGNAL excep,
				  uint16 addr, const char *message);

#undef inline
#define inline static __inline__


inline uint8
memory_read8 (sim_cpu * cpu, uint16 addr)
{
  uint8 val;

  if (sim_core_read_buffer (CPU_STATE (cpu), cpu, 0, &val, addr, 1) != 1)
    {
      cpu_memory_exception (cpu, SIM_SIGSEGV, addr, "Read error");
    }
  return val;
}

inline void
memory_write8 (sim_cpu * cpu, uint16 addr, uint8 val)
{
  if (sim_core_write_buffer (CPU_STATE (cpu), cpu, 0, &val, addr, 1) != 1)
    {
      cpu_memory_exception (cpu, SIM_SIGSEGV, addr, "Write error");
    }
}

inline uint16
memory_read16 (sim_cpu * cpu, uint16 addr)
{
  uint8 b[2];

  if (sim_core_read_buffer (CPU_STATE (cpu), cpu, 0, b, addr, 2) != 2)
    {
      cpu_memory_exception (cpu, SIM_SIGSEGV, addr, "Read error");
    }
  return (((uint16) (b[0])) << 8) | ((uint16) b[1]);
}

inline void
memory_write16 (sim_cpu * cpu, uint16 addr, uint16 val)
{
  uint8 b[2];

  b[0] = val >> 8;
  b[1] = val;
  if (sim_core_write_buffer (CPU_STATE (cpu), cpu, 0, b, addr, 2) != 2)
    {
      cpu_memory_exception (cpu, SIM_SIGSEGV, addr, "Write error");
    }
}

inline void
check_stack (sim_cpu * proc, int newoffset)
{
  if (proc->cpu_regs.sp + newoffset + 1 < 0xC0 ||
      proc->cpu_regs.sp + newoffset > 0xff)
    {
      cpu_set_pc (proc, proc->cpu_insn_pc);
      sim_engine_halt (CPU_STATE (proc), proc, NULL,
		       cpu_get_pc (proc), sim_stopped, SIM_SIGSEGV);
    }

}


/* Fetch a 8/16 bit value and update the PC.  */
inline uint8
cpu_fetch8 (sim_cpu * proc)
{
  uint16 addr = proc->cpu_regs.pc;
  uint8 val;

  val = memory_read8 (proc, addr);
  proc->cpu_regs.pc = addr + 1;
  return val;
}

inline uint16
cpu_fetch16 (sim_cpu * proc)
{
  uint16 addr = proc->cpu_regs.pc;
  uint16 val;

  val = memory_read16 (proc, addr);
  proc->cpu_regs.pc = addr + 2;
  return val;
}

inline void
cpu_push_uint8 (sim_cpu * proc, uint8 val)
{
  uint16 addr = proc->cpu_regs.sp;

  check_stack (proc, -1);
  memory_write8 (proc, addr, val);
  proc->cpu_regs.sp = addr - 1;
}

inline void
cpu_push_uint16 (sim_cpu * proc, uint16 val)
{
  uint16 addr = proc->cpu_regs.sp - 1;

  check_stack (proc, -2);
  memory_write16 (proc, addr, val);
  proc->cpu_regs.sp = addr - 1;
}

inline uint8
cpu_pop_uint8 (sim_cpu * proc)
{
  uint16 addr = proc->cpu_regs.sp;
  uint8 val;

  check_stack (proc, 1);
  val = memory_read8 (proc, addr + 1);
  proc->cpu_regs.sp = addr + 1;
  return val;
}

inline uint16
cpu_pop_uint16 (sim_cpu * proc)
{
  uint16 addr = proc->cpu_regs.sp;
  uint16 val;

  check_stack (proc, 2);
  val = memory_read16 (proc, addr + 1);
  proc->cpu_regs.sp = addr + 2;
  return val;
}

inline void
cpu_ccr_update_NZ (sim_cpu * proc, uint8 val)
{
  cpu_set_ccr_Z (proc, val == 0);
  cpu_set_ccr_N (proc, val & 0x80);
}
inline void
cpu_ccr_update_NZC (sim_cpu * proc, uint16 val)
{
  cpu_set_ccr_Z (proc, val == 0);
  cpu_set_ccr_N (proc, val & 0x80);
  cpu_set_ccr_C (proc, val & 0x100);
}
#endif
